"""Module containing a parent class for an azol http client"""
import uuid
import logging
import requests
from azol.utils import is_token_expired, get_tenant_id, parse_jwt
from azol.services.token_service import TokenService as tService
from azol.credentials import User

class OAuthHTTPClient:
    """
        Base class for OAuth HTTP client objects.
    """

    def __init__( self, cred, oauth_resource, base_url,
                  tenant=None, azol_id=None, oauth_flow=None, secrets_provider=None,
                  use_persistent_cache=True, auto_refresh=True, scopes=[]):
        if tenant is None:
            if cred.credentialType != "user":
                raise Exception("tenant must be specified if credential is not of type 'user'")
            if not cred.default_oauth_flow=="refresh_token":
                username=cred.get_username()
                tenant=username.split("@")[1]
        self.scopes=scopes
        self.default_scope=True
        self.profile_scope=True
        self.openid_scope=True
        self.offline_access_scope=True
        self._baseurl = base_url

        self._tenant = tenant
        self._resource = oauth_resource
        self._current_token = None

        # if no specific oauth flow is provided, get the default from the credential
        if oauth_flow is None:
            oauth_flow=cred.get_default_oauth_flow()

        self.token_service = tService( cred, tenant_id=tenant, oauth_flow=oauth_flow,
                                    secrets_provider=secrets_provider,
                                    use_persistent_cache=use_persistent_cache,
                                    oauth_resource=oauth_resource, scopes=self.scopes,
                                    default_scope=self.default_scope,
                                    profile_scope=self.profile_scope,
                                    openid_scope=self.openid_scope,
                                    offline_access_scope=self.offline_access_scope )

        self._current_token = self.token_service.get_cached_token_if_not_expired()

        self._auto_refresh=auto_refresh
        if azol_id is None:
            azol_id = str(uuid.uuid4())
        self._id=azol_id

    def get_token_claims( self ):
        """Get claims in the current token

        parses the cached token and returns the body

        Returns:
           A dictionary containing the unencoded token body, or None if no token

        """
        token = self.get_current_token()

        if token is None:
            return None
        
        header, body, signature = parse_jwt(token)

        return body

    def switch_tenant(self, tenant):
        """
            Switch the tenant of the client.
            This nullifies the current token of the client, and subsequent requests
            will be made to the new tenant

            Args:
                - tenant - (string)  a tenant id or domain name for a tenant
            Returns:
                None
        """
        # get the tenant Id if its a domain name. If it cant be parsed as a guid, its a domain name.
        try:
            uuid.UUID(tenant, version=4)
        except ValueError:
            tenant=get_tenant_id(tenant)

        self._current_token=None
        self.token_service.set_tenant(tenant)
        self._tenant=tenant

    def switch_scope(self, scopes, default, profile, offline_access, openid):
        """
            Change the scope of the OAuth client. This will cause a new scope to be requested
            on subsequent tokens, and will nullify the current token.

            Args:
                - scopes - a list containing a list of oauth scopes to request
                - default - bool, if the default scope should be requested. 
                              this overrides any other scope set
                - profile - bool, if the profile scope should be requested
                - offline_access - bool, if the offline_access scope should be requested
                - openid - bool, if the openid scope should be requested

            Returns:
                None
        """
        self._current_token=None
        self.token_service.set_scope(scopes, default, profile, offline_access, openid)
        self.scopes=scopes
        self.default_scope=default
        self.profile_scope=profile
        self.offline_access_scope=offline_access
        self.openid_scope=openid

    def get_id(self):
        """
            Get the azol ID of the client

            Returns:
                 string
        """
        return self._id

    def refresh_token( self ):
        """
            force the token cache to use the refresh token in the current credential to
            generate a new access token

            Returns:
                None
        """
        self._current_token = self.token_service.refresh_token()

    def get_current_refresh_token( self ):
        """
            Get the current refresh token in use by the client

            Returns:
                string - ascii representation of refresh token
        """
        token=self.token_service.get_refresh_token_if_exists()
        return token

    def fetch_token(self):
        """
            Fetch a new token. Save the token as the current token to be used by the client

            Args:
                - scope - (string) a space-separated list of scopes to request for the new token

            Returns:
                None
        
        """
        # token service needs to "log in" and cache the token.
        # if the credential is a user credential, also possible cache the refresh token
        access_token=self.token_service.fetch_token()
        self._current_token = access_token
        return None

    def get_current_token(self):
        """
            Get the current token being used by the client

            Returns:
                raw ascii representation of the current token or None if no token is saved
        """
        if self._current_token is None:
            return None

        return self._current_token

    def get_client_id( self ):
        return self.token_service.get_client_id()

    def switch_client(self, client_id):
        """
            Switch the oauth client used by the HTTP Client. Used to abuse FOCI functionality.

            nullifies current cached access token, but leaves refresh token in client.

            Vars:
                - client_id - the client id of the client that we will switch to.

            Returns:
                OAuthHTTPClient subclass
        """
        self._current_token = None
        self.token_service.switch_client(client_id)

    @classmethod
    def from_client(cls, client, *args, **kwargs):
        """
            Generate a new HTTP Client with a different OAuth Resource, using the current client class.

            Use this class to switch from, for example, a GraphClient to an ArmClient.
            Nullifies current cached access token, but leaves refresh token in client.

            Note: alternative to "refresh_to_new_resource" of an active client, but results in the same.

            Vars:
                - client - OAuthHTTP child class that will be used to instantiate the 
                              new class. Note that this must be a valid
                              OAuth HTTP child class. For example, ArmClient, GraphClient,
                              or KeyVaultClient

            Returns:
                OAuthHTTPClient subclass instance
        """
        rt = client.get_current_refresh_token()
        user = User(refresh_token=rt, client_id=client.get_client_id())
        newClass = cls(*args, tenant=client._tenant, cred=user, **kwargs)
        return newClass

    def refresh_to_new_resource(self, oauth_http_client_class):
        """
            Generate a new HTTP Client with a different OAuth Resource, using the current token.

            Use this class to switch from, for example, a GraphClient to an ArmClient.
            Nullifies current cached access token, but leaves refresh token in client.

            Vars:
                - oauth_http_client_class - OAuthHTTP child class to instantiate with the current
                              class's refresh token. Note that this must be a valid
                              OAuth HTTP child class. For example, ArmClient, GraphClient,
                              or KeyVaultClient

            Returns:
                OAuthHTTPClient subclass instance
        """
        rt = self.get_current_refresh_token()
        user = User(refresh_token=rt)
        client = oauth_http_client_class(tenant=self._tenant, cred=user)
        return client

    def _set_current_token(self, token):
        """
            internal: Set the current token

            Returns:
                None
        """
        self._current_token = token

    def _send_request( self, path=None, query_parameters=None, data=None, method="GET",
                       json=None, url=None ):
        """
            Internal method. Send request
        """
        if query_parameters is None:
            query_parameters={}
        if url is not None:
            request_url = url
        else:
            request_url=f"{self._baseurl}{path}"
        #current token is the token that was last set when calling "fetchAndCacheToken"
        access_token = self.get_current_token()

        if access_token is None and self._auto_refresh is False:
            logging.error("cannot sent request because credential is not logged in yet."
                          "call fetchAndCacheToken first")
            raise Exception("Auto Refresh is turned off and there is no cached token")
        elif (access_token is not None
              and self._auto_refresh
              and is_token_expired( self._current_token, 600 ) ):
            # if token expires in the next 10 minutes, go get a new one and cache it locally.
            self.fetch_token()
        elif access_token is None and self._auto_refresh is True:
            self.fetch_token()

        access_token = self.get_current_token()
        headers = {'Authorization': f"Bearer {access_token}" }
        resp = requests.request( method, request_url, data=data, params=query_parameters,
                                 headers=headers, json=json, timeout=10)

        return resp



"""A module containing the parent class for an EntraID Credential"""
from azol.credentials.credential import Credential

class EntraIdCredential( Credential ):
    """
        Parent class of other Entra ID credential classes. Sub-classes simply overwrite "getToken"
    """

    supportedOAuthFlows = []
    credentialType=None
    default_oauth_flow=None
    def __init__( self, *args, **kwargs ):
        super().__init__( *args, **kwargs)
    


"""A module containing an azol key vault http client"""
import logging
from azol.clients.oauth_http_client import OAuthHTTPClient
from azol.constants import KEYVAULTAPIVERSION, OAuthResourceIDs

class KeyVaultClient( OAuthHTTPClient ):
    """
        An HTTP client for interacting with Azure key vaults.
        Works with all credential objects
    """

    def __init__( self, key_vault_name, *args, **kwargs ):
        key_vault_resource_id=OAuthResourceIDs.KeyVault
        kv_base_url=f"https://{key_vault_name}.vault.azure.net"
        super().__init__( oauth_resource=key_vault_resource_id,
                          base_url=kv_base_url, *args, **kwargs)
        self.kv_name = key_vault_name

    def get_keys( self ):
        """Get all keys in the key vault.

        Get all keys from the key vault connected to this client,
        which the credential object can access.

        Returns:
           None if error, or <dict> containing raw response from the key vault
        """
        query_params = { "api-version": KEYVAULTAPIVERSION}
        raw_response = self._send_request( "/keys", query_parameters=query_params )

        if raw_response.status_code != 200:
            logging.error( "ERROR while getting keys from key vault" )
            logging.error( raw_response.json())
            return None
        return raw_response.json()['value']

    def get_secret( self, secret_name, secret_version=None ):
        """Get a secret from the key vault.

        Args:
            - secret_name - (string) The name of the secret
            - secret_version - (string) defaults to None. the version of the secret, or 
                               None if requesting the most recent
        
        Returns:
           None if error, or <dict> containing raw response from the key vault
        """
        query_params = { "api-version": KEYVAULTAPIVERSION}
        
        if secret_version is None:
            raw_response = self._send_request( f"/secrets/{secret_name}",
                                             query_parameters=query_params )
        else:
            raw_response = self._send_request( f"/secrets/{secret_name}/{secret_version}",
                                             query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting secret from key vault" )
            logging.error( raw_response.json())
            return None
        return raw_response.json()['value']

    def get_secrets( self ):
        """Get all secrets in the key vault.

        Get all secrets from the key vault connected to this client,
        which the credential object can access.

        Returns:
           None if error, or <dict> containing raw response from the key vault
        """
        query_params = { "api-version": KEYVAULTAPIVERSION}
        raw_response = self._send_request( "/secrets", query_parameters=query_params )

        if raw_response.status_code != 200:
            logging.error( "ERROR while listing secrets from key vault" )
            logging.error( raw_response.json())
            return None
        return raw_response.json()['value']

    def get_certificates( self ):
        """Get all certificates in the key vault.

        Get all certificates from the key vault connected to this client,
        which the credential object can access.

        Returns:
           None if error, or <dict> containing raw response from the key vault
        """
        query_params = { "api-version": KEYVAULTAPIVERSION}
        raw_response = self._send_request( "/certificates", query_parameters=query_params )

        if raw_response.status_code != 200:
            logging.error( "ERROR while getting cert from key vault" )
            logging.error( raw_response.json())
            return None
        return raw_response.json()['value']



from setuptools import setup, find_packages
import os

version = os.environ.get('BUILD_VERSION')

readmeText=""
with open("README.md", "r") as f:
    readmeText+=f.read()

setup(
    name='azol',
    version=version,
    packages=find_packages(),
    url='https://github.com/cdburkard/azol',
    author='Cody Burkard',
    description='A python-based pentesting library for Azure and Entra ID',
    long_description=readmeText,
    long_description_content_type="text/markdown",
    install_requires=[
        "requests==2.31.0",
        "dataclasses==0.6"
    ]
)


"""A module containing the Application Object credential class"""
from azol.credentials.entraid_credential import EntraIdCredential
from azol.constants import OAUTHFLOWS

class ApplicationObject( EntraIdCredential ):
    """
        A credential that may be used when a client secret and 
        client id of an application object is known
    """

    supportedOAuthFlows = [ OAUTHFLOWS.CLIENT_CREDENTIALS ]
    credentialType="app"
    default_oauth_flow=OAUTHFLOWS.CLIENT_CREDENTIALS

    def __init__( self, client_id, client_secret, *args, **kwargs ):
        super().__init__( *args, **kwargs)
        self._client_id = client_id
        self._client_secret = client_secret

    def get_client_secret( self ):
        """Get the client secret that is saved in this credential class

            Returns: A string containing the client secret
        """
        return self._client_secret



"""A module containing a client for interacting with the Graph API"""
import logging
from azol.clients.oauth_http_client import OAuthHTTPClient
from azol.constants import GRAPHBETAURL, OAuthResourceIDs, roleNameMap, appPermissionNameMap

class GraphRequestFailedException(Exception):
    """
        Exception that is raised when requests to MS Graph
        unexpectedly fail
    """

class GraphClient( OAuthHTTPClient ):
    """
        An HTTP client for interacting with the Microsoft Graph API
    """

    def __init__( self, *args, base_url=GRAPHBETAURL, **kwargs ):
        super().__init__(  oauth_resource=OAuthResourceIDs.Graph, base_url=base_url, *args, **kwargs )

    def _send_request( self, *args, success_code=200, **kwargs ):
        response = super()._send_request(*args, **kwargs)
        if response.status_code != success_code:
            logging.error( "Error on GRAPH API request. Raw error: %s ", str(response.content))
            raise GraphRequestFailedException()
        return response

    def _get_all_graph_objects(self, response):
        """Get all objects from graph api request

        read all pages from a graph api request
        
        Args:
            response - (requests.Response) TThe requests response object for a graph request

        Returns:
            A list of dictionaries containing the entire Graph response

        """

        all_objects=[]
        objects=response.json()["value"]
        all_objects += objects
        while "@odata.nextLink" in response.json().keys():
            response = self._send_request( url=response.json()["@odata.nextLink"] )
            objects = response.json()[ "value" ]
            all_objects += objects
        return all_objects

    def get_directory_role_definitions( self ):
        """Get all directory role definitions.
        
        Returns:
            A list of dictionaries containing the role definitions in the directory.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response = self._send_request( "/roleManagement/directory/roleDefinitions?"
                                       "$select=displayName,id,isBuiltIn" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_access_catalogs_roles( self, catalogId ):
        """
            Get all the roles assigned to a specific access catalog

            Args:
                catalogId - the ID of the catalog

            Returns:
                A list of dictionaries containing role assignments for the catalog

            Raises:
                GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response=self._send_request( f"/roleManagement/entitlementManagement/roleAssignments?$filter=appScopeId eq '/AccessPackageCatalog/{catalogId}'" )

        if response:
            roles=response.json()["value"]
            for role in roles:
                role["azolAnnotations"] = {
                    "roleName" : roleNameMap[role["roleDefinitionId"]]
                }
            return roles
        raise GraphRequestFailedException()


    def create_package_policy( self, policy ):
        """Create policy for Entitlement Management Pacakge

        Returns:
            A dictionary containing the entitlement management package policy.

        Args:
            catalog - dictionary containing entitlement management package policy

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response=self._send_request( "/identityGovernance/entitlementManagement/accessPackageAssignmentPolicies",
                                     method="POST", json=policy, success_code=201 )

        if response:
            return response.json()
        raise GraphRequestFailedException()

    def create_entitlement_management_package( self, package ):
        """Create Entitlement Management Pacakge

        Returns:
            A dictionary containing the entitlement management package.

        Args:
            catalog - dictionary containing entitlement management package metadata

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response=self._send_request( "/identityGovernance/entitlementManagement/accessPackages",
                                     method="POST", json=package, success_code=201 )

        if response:
            return response.json()
        raise GraphRequestFailedException()

    def create_entitlement_management_catalog( self, catalog ):
        """Create Entitlement Management Catalog

        Returns:
            A list of dictionaries containing all entitlement management catalog.

        Args:
            catalog - dictionary containing entitlement management catalog metadata

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response=self._send_request( "/identityGovernance/entitlementManagement/accessPackageCatalogs",
                                     method="POST", json=catalog, success_code=201 )

        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_access_packages( self ):
        """Get Entitlement Management Access Packages.

        Returns:
            A list of dictionaries containing all entitlement management access packages metadata.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """

        response=self._send_request( "/identityGovernance/entitlementManagement/accessPackages?$expand=accessPackageCatalog($select=displayName,id,description)" )

        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_entitlement_management_catalogs( self ):
        """Get Entitlement Management Catalogs.

        Returns:
            A list of dictionaries containing entitlement management catalog metadata.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """

        response=self._send_request( "/identityGovernance/entitlementManagement/accessPackageCatalogs" )

        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_current_pim_eligibility( self ):
        """Get pim eligibility of the current principal.

        Returns:
            A list of dictionaries containing the pim eligibility of the current principal.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response = self._send_request( "/roleManagement/directory/roleEligibilitySchedules/"
                                       "filterByCurrentUser(on='principal')?"
                                       "$expand=principal,roleDefinition($select=displayName,id,templateId)"
                                       "&$select=principal,roleDefinition,scheduleInfo,memberType,appScopeId,"
                                       "directoryScopeId,createdDateTime" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_current_pim_activations( self ):
        """Get pim activations of the current principal.

        Returns:
            A list of dictionaries containing the pim activations of the current principal.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response = self._send_request( "/roleManagement/directory/roleAssignmentSchedules/"
                                       "filterByCurrentUser(on='principal')?"
                                       "$expand=principal,roleDefinition($select=displayName,id,templateId)"
                                       "&$select=principal,roleDefinition,scheduleInfo,memberType,appScopeId,"
                                       "directoryScopeId,createdDateTime" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_active_pim_assignments( self ):
        """Get all active PIM sessions.
        
        Returns:
            A list of dictionaries containing the active PIM sessions in the directory.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response = self._send_request( "/roleManagement/directory/roleAssignmentSchedules?"
                                       "$expand=principal,roleDefinition($select=displayName,id,templateId)"
                                       "&$select=principal,roleDefinition,scheduleInfo,memberType,appScopeId,"
                                       "directoryScopeId,createdDateTime" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_eligible_pim_assignments( self ):
        """Get all eligible PIM assignments.
        
        Returns:
            A list of dictionaries containing the eligible PIM assignments in the directory.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response = self._send_request( "/roleManagement/directory/roleEligibilitySchedules?"
                                       "$expand=principal,roleDefinition($select=displayName,id,templateId)"
                                       "&$select=principal,roleDefinition,scheduleInfo,memberType,appScopeId,"
                                       "directoryScopeId,createdDateTime" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_transitive_group_memberships( self, group_id ):
        """Get transitive group memberships of a specific group.
        
        Args:
            group_id - (str) object id of the group

        Returns:
            A list of objects containing members of the group.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API

        """
        response = self._send_request( f"/groups/{group_id}/transitiveMembers" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_users( self, select=[] ):
        """Get all users in the directory.

        Args:
            select - (list) - a list of json attributes that should be returned from all group.
                    if this is None, the select query parameter will not be used in the request
        Returns:
            A list of objects containing object id and displayName of all users

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        if select == []:
            path_and_params="/users?$select=id,displayName"
        elif select == None:
            path_and_params="/users"
        else:
            path_and_params="/users?$select="+",".join(select)
        response = self._send_request( path_and_params )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_service_principals( self, select=[] ):
        """Get all service principals in the directory.

        Args:
            select - (list) - a list of json attributes that should be returned from all group.
                    if this is None, the select query parameter will not be used in the request
        Returns:
            A list of objects containing object id and displayName of all service principals

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        if select == []:
            path_and_params="/servicePrincipals?$select=id,displayName"
        elif select == None:
            path_and_params="/servicePrincipals"
        else:
            path_and_params="/servicePrincipals?$select="+",".join(select)
        
        response = self._send_request( path_and_params )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_applications( self, select=[] ):
        """Get all application objects in the directory.

        Returns:
            A list of objects containing object id and displayName of all application objects

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        if select == []:
            path_and_params="/applications?$select=id,displayName"
        elif select == None:
            path_and_params="/applications"
        else:
            path_and_params="/applications?$select="+",".join(select)
        response = self._send_request( path_and_params )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_directory_role_assignments( self, object=None ):
        """Get directory role assignments.

        Get all Entra ID role assignments in the directory, ignoring assignments through PIM

        Returns:
            A list of dictionaries containing role assignment metadata

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        if object is None:
            response = self._send_request( "/roleManagement/directory/roleAssignments?$expand=roleDefinition($select=id,displayName)&select=roleDefinition,roleDefinitionId,principalId,resourceScope,principalOrganizationId" )
        else:
            response = self._send_request( f"/roleManagement/directory/roleAssignments?$expand=roleDefinition($select=id,displayName)&select=roleDefinition,roleDefinitionId,principalId,resourceScope,principalOrganizationId&$count=true&$filter=principalId eq '{object}'" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def delete_directory_role_assignments( self, assignment_id ):
        """Remove an Entra ID role from a principal in the directory.

        Args:
            assignment_id - (str) - The ID of the role assignment to delete

        Returns:
            A list of dictionaries containing the graph response to the deletion

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/roleManagement/directory/roleAssignments/{assignment_id}",
                                       method="DELETE" )
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def add_directory_role_assignment( self, role_definition_id, principal_id ):
        """Assign an Entra ID role to a principal in the directory.

        Args:
            role_definition_id - (str) - The ID of the role to assign
            principal_id - (str) - the principal to assign to the role

        Returns:
            A dictionary containing the graph ressponse

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        body = {
            "@odata.type": "#microsoft.graph.unifiedRoleAssignment",
            "roleDefinitionId": role_definition_id,
            "principalId": principal_id,
            "directoryScopeId": "/"
        }
        response = self._send_request( "/roleManagement/directory/roleAssignments",
                                       method="POST", success_code=201, json=body )
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def add_app_owner( self, app_object_id, principal_id ):
        """Add an owner to an application object.

        Args:
            app_object_id - (str) - The ID of the application
            principal_id - (str) - the principal to assign to owner

        Returns:
            A dictionary containing the graph ressponse

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        body = {
            "@odata.id": f"https://graph.microsoft.com/v1.0/directoryObjects/{principal_id}"
        }
        # ERROR: this always errors out, even if it succeeds. its a requests library error
        response = self._send_request( f"/applications/{app_object_id}/owners/$ref",
                                       method="POST", json=body )

        if response:
            return response.json()
        raise GraphRequestFailedException()

    def remove_app_owner( self, app_object_id, principal_id ):
        """Remove an owner from an application object.

        Args:
            app_object_id - (str) - The ID of the application
            principal_id - (str) - the principal to assign to owner

        Returns:
            A dictionary containing the graph ressponse

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/applications/{app_object_id}/owners/{principal_id}/$ref",
                                       method="DELETE")
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def get_all_service_principals_owners( self ):
        """Get all owners of all service principals.

        Calls the graph API get get all service prinicpals, as well as their
        owners.

        Returns:
            A list of dictionaries, containing service principal ids,
            names, and owners

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( "/servicePrincipals?$expand=owners($select=id,displayName)&"
                                       "$select=owners,id,displayName" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_application_owners( self ):
        """Get all owners of all application objects.

        Calls the graph API get get all application objects, as well as their
        owners.

        Returns:
            A list of dictionaries, containing app object ids,
            names, and owners

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( "/applications?$expand=owners($select=id,displayName)&"
                                       "$select=owners,id,displayName" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_groups_and_memberships( self ):
        """Get all groups and group memberships.

        Calls the graph API get get all groups, and nested memberships.

        Returns:
            A list of dictionaries, containing group ids, names, and nested relationships

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( "/groups?$expand=memberOf($select=id,displayName)"
                                       "&$select=memberOf,id,displayName" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_groups( self, select=[] ):
        """Get all group ids and displaynames in Entra Id.

        Calls the graph API get get all groups, and nested memberships.

        Args:
            select - (list) - a list of json attributes that should be returned from all group.
                    if this is None, the select query parameter will not be used in the request

        Returns:
            A list of dictionaries, containing group ids and names

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        if select == []:
            path_and_params="/groups?$select=id,displayName"
        elif select == None:
            path_and_params="/groups"
        else:
            path_and_params="/groups?$select="+",".join(select)
        response = self._send_request( path_and_params )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_groups_and_owners( self ):
        """Get all group owners in all Entra ID groups.

        Calls the graph API get get all groups, and owners of those groups.

        Returns:
            A list of dictionaries, containing groupIds, groupNames, and owners

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( "/groups?$expand=owners($select=id,displayName)"
                                       "&$select=owners,id,displayName" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_all_sp_api_permissions( self ):
        """Get all the API permissions assigned to all service principals.

        Returns:
           A list of dictionaries with API permissions assigned to all service principal

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( "/servicePrincipals?$expand=appRoleAssignments("
                                       "$select=resourceId,resourceDisplayName,principalType,"
                                       "appRoleId)&$select=appRoleAssignments,id,displayName" )
        if response:
            sps = self._get_all_graph_objects(response)
            for sp in sps:
                for ass in sp["appRoleAssignments"]:
                    from pprint import pprint
                    ass["azolannotations"] = {
                        "permissionName": appPermissionNameMap[ass["appRoleId"]]
                    }
            return sps
        raise GraphRequestFailedException()

    #def get_all_sp_delegated_permissions( self ): # TODO: currently errors
    #    """Get all the API permissions assigned to all service principals.
    #
    #    Returns:
    #       A list of dictionaries with API permissions assigned to all service principal
    #
    #    Raises:
    #        GraphRequestFailedException: An error occurred accessing the Graph API
    #    """
    #    response = self._send_request( "/servicePrincipals?&$select=oauth2PermissionGrants,"
    #                                   "id,displayName" )
    #    if response:
    #        return self._get_all_graph_objects(response)
    #    raise GraphRequestFailedException()

    def get_service_principal(self, sp_object_id):
        """Get service principals.

        Returns:
           A list of dictionaries containing service principal properties from Graph

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/servicePrincipals/{sp_object_id}" )
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def get_all_sp_reply_urls( self ):
        """Get all the reply Urls for all service principals.

        Returns:
           A list of dictionaries containing service principals and
           their reply urls.

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( "/servicePrincipals?&$select=replyUrls,id,displayName" )
        if response:
            return self._get_all_graph_objects(response)
        raise GraphRequestFailedException()

    def get_api_permissions( self, sp_object_id ):
        """Get all API permissions assigned to a service principal.

        Get all the API permissions assigned to a service principal. 

        Args:
            - sp_object_id - (string) The object ID of a service principal

        Returns:
           A dictionary containing the service principal and
           its API permissions 

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/servicePrincipals/{sp_object_id}/appRoleAssignments" )
        if response:
            api_permissions=self._get_all_graph_objects(response)
            for permission in api_permissions:
                permission["azolAnnotations"] = {
                        "permissionName": appPermissionNameMap[permission["appRoleId"]]
                }
            return api_permissions
        raise GraphRequestFailedException()

    def get_delegated_permissions( self, sp_object_id ):
        """Get all the delegated permissions assigned to a service principal.

        Get all the API permissions assigned to a service principal. 

        Includes all users that have consented to the permission.
        Principal Id displayNames are annotated at runtime for human readability.

        Args:
            - sp_object_id - (string) The object ID of a service principal

        Returns:
           A dictionary containing the service principal and
           its consented delegated permissions 

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/servicePrincipals/{sp_object_id}/oauth2PermissionGrants" )
        if response.status_code != 200:
            logging.error( "error on GRAPH API request. Raw error: %s", response.content )
            raise GraphRequestFailedException()

        delegated_permissions=response.json()["value"]
        annotated_permissions=[]
        for p in delegated_permissions:
            annotated_permission = p.copy()
            if p["consentType"] == "AllPrincipals":
                annotated_permission["principalName"] = "all"
            else:
                resp = self._send_request( f"/directoryObjects/{p[ 'principalId' ]}" )
                obj = resp.json()
                principal_name = obj["userPrincipalName"]
                annotated_permission[ "principalName" ] = principal_name
            annotated_permissions.append(annotated_permission)
        return annotated_permissions

    def get_all_principals( self ):
        """Get all principals in the directory.

        Get all users, service principals, and groups in the directory.

        These are the objects that may be assigned permissions in M365 and Azure. 

        Returns:
           A dictionary mapping object ids to users, groups and service principals

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        users=self.get_all_users()
        service_principals=self.get_all_service_principals()
        groups=self.get_all_groups()
        principals = {}
        for directory_object in users:
            val = {
                "displayName": directory_object["displayName"],
                "principalType": "User"
            }
            key = directory_object["id"]
            principals[key] = val

        for directory_object in service_principals:
            val = {
                "displayName": directory_object["displayName"],
                "principalType": "ServicePrincipal"
            }
            key = directory_object["id"]
            principals[key] = val

        for directory_object in groups:
            val = {
                "displayName": directory_object["displayName"],
                "principalType": "Group"
            }
            key = directory_object["id"]
            principals[key] = val

        return principals

    def add_app_secret( self, app_object_id, name="inconspicuous" ):
        """Add a secret to a service principal in Entra ID.

        Args:
            - app_object_id - (string) the object ID of the Entra ID
                                application for which to generate a new secret
            - name - (string) The name of the new secret.

        Returns:
           A dictionary containing the properties for the new app secret

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        body= {
            "passwordCredential": {
                "displayName": name
            }
        }
        response = self._send_request( f"/applications/{app_object_id}/addPassword",
                                      method="POST", json=body )
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def add_sp_secret( self, sp_object_id, name="inconspicuous" ):
        """Add a secret to an application object in Entra ID.

        Args:
            - sp_object_id - (string) the object ID of the Entra ID
                            service principal for which to generate a new secret.
            - name - (string) The name of the new secret.

        Returns:
           A dictionary containing the properties for the new sp secret

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        body= {
            "passwordCredential": {
                "displayName": name
            }
        }
        response = self._send_request( f"/servicePrincipals/{sp_object_id}/addPassword",
                                       method="POST", json=body )
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def get_directory_object( self, object_id ):
        """Get a directory object from its object id.

        Attempt to get a directory object from its object id. This could be any type
        of directory object. Useful when a guid is identified, and it is not clear what it is.

        Args:
            - object_id - (string) The directory object Id.

        Returns:
           A dictionary containing information from graph about the object

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/directoryObjects/{object_id}" )

        if response:
            directory_object = response.json()
            return directory_object
        raise GraphRequestFailedException()

    def try_get_object_type( self, object_id ):
        """Get a directory object type from its object id.

        Attempt to get a directory object's type base on its object id. This could be any type
        of directory object. Useful when a guid is identified, and it is not clear what it is.

        Args:
            - object_id - (string) The directory object Id.

        Returns:
           (string) the type of the directory object, from Graph, or None if it is not a
                    directory object

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( f"/directoryObjects/{object_id}" )

        if response:
            if "@odata.type" in response.json().keys():

                directory_object_type = response.json()[ "@odata.type" ]
                return directory_object_type
            return None
        raise GraphRequestFailedException()

    def create_new_local_service_principal( self, name="inconspicuous" ):
        """Create a new service principal.

        Create an application object an service principal in the current tenant of the client.
        Requires a user with an Entra ID role or a service principal with a graph API
        permission capable of creating service principals.

        Also creates a secret for this new service principal.

        Args:
            - name - (string) The name of the new service principal.

        Returns:
           (string) A dictionary containing properties for the new service principal,
                    including the newly created secret

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        body = {
            "displayName": name
        }
        response = self._send_request( "/applications", method="POST", json=body )

        if response.status_code != 201:
            logging.error( "error on GRAPH API request. Raw error: %s ", response.content )
            raise GraphRequestFailedException()

        app = response.json()
        app_id = app[ "appId" ]
        app_object_id = app[ "id" ]

        # Create the logcal service principal
        body = {
            "appId": app_id
        }
        response = self._send_request( "/servicePrincipals", method="POST", json=body )

        if response.status_code != 201:
            logging.error( "error on GRAPH API request. Raw error: %s ", response.content )
            raise GraphRequestFailedException()

        sp = response.json()
        sp_id = sp[ "id" ]

        # create a secret for the service principal
        body = {
            "passwordCredential": {
                "displayName": "inconspicuous"
            }
        }
        response = self._send_request( f"/servicePrincipals/{sp_id}/addPassword",
                                      method="POST", json=body )

        if response.status_code != 200:
            logging.error( "error on GRAPH API request. Raw error: %s ", response.content )
            raise GraphRequestFailedException()

        secret = response.json()[ "secretText" ]

        output = {
            "clientId": app_id,
            "appObjectId": app_object_id,
            "spId": sp_id,
            "spSecret": secret
        }

        return output

    def create_new_remote_service_principal( self, client_id ):
        """Create a new service principal referencing a remote application.

        Create avservice principal in the current tenant of the client.
        This service principall will reference an application in another directory.

        This means the application object in the other directory may be used to access
        the permissions of this principal

        Requires a user with an Entra ID role or a service principal with a graph API
        permission capable of creating service principals.

        Args:
            - client_id - (string) The client id of the service principal.

        Returns:
           (string) A dictionary containing properties for the new service principal

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        # Create the service principal
        body = {
            "appId": client_id
        }
        response = self._send_request( "/servicePrincipals", method="POST", json=body, success_code=201 )

        sp = response.json()
        sp_id = sp[ "id" ]

        output = {
            "clientId": client_id,
            "spId": sp_id
        }

        return output

    def get( self, path ):
        """Make a get request to a specific API path within the Graph API.

        Args:
            - path - (string) The graph API path.

        Returns:
           Request.Response object from the GET request

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( path )
        if response:
            return response.json()
        raise GraphRequestFailedException()

    def post( self, path, data ):
        """Send a POST request to Graph, and return the raw results

        Calls Graph at the path specified, and returns the raw deserialized results.
        
        Args:
            path - (str) The API path to call on Graph
            data - (dict) The JSON data to send to Graph in the API request body

        Returns:
            A dictionary containing the raw results from the ARM request.

        """
        response = self._send_request( path, method="POST",
                                              json=data )
        return response.json()

    def put( self, path, data ):
        """Send a PUT request to Graph, and return the raw results

        Calls Graph at the path specified, and returns the raw deserialized results.
        
        Args:
            path - (str) The API path to call on Graph
            data - (dict) The JSON data to send to Graph in the API request body

        Returns:
            A dictionary containing the raw results from the ARM request.

        """
        response = self._send_request( path, method="PUT",
                                              json=data )
        return response.json()

    def patch( self, path, data ):
        """Send a PATCH request to Graph, and return the raw results

        Calls Graph at the path specified, and returns the raw deserialized results.
        
        Args:
            path - (str) The API path to call on Graph
            data - (dict) The JSON data to send to Graph in the API request body

        Returns:
            A dictionary containing the raw results from the ARM request.

        """
        response = self._send_request( path, method="PATCH",
                                              json=data )
        return response.json()

    def delete( self, path ):
        """Make a DELETE request to a specific API path within the Graph API.

        Args:
            - path - (string) The graph API path.

        Returns:
           Request.Response object from the DELETE request

        Raises:
            GraphRequestFailedException: An error occurred accessing the Graph API
        """
        response = self._send_request( path, method="DELETE" )
        if response:
            return response.json()
        raise GraphRequestFailedException()



"""A module containing the parent credential class"""
import uuid

class Credential:
    """
        Parent class of other Entra ID credential classes. Sub-classes simply overwrite "getToken"
    """

    credentialType=None
    supportedOAuthFlows = []
    default_oauth_flow=None
    def __init__( self, azol_id=None ):

        self._client_id=None
        self._username=None
        if azol_id is None:
            azol_id = uuid.uuid4()
        self._id = str(azol_id)

    def get_default_oauth_flow( self ):
        """
            Get the default oauth flow for the credential type.
            This oauth flow is one of the following:
            - client_credential
            - raw_token
            - refresh_token
            - device_code

            This is defined in each credential class in the default_oauth_flow class variable

            Returns: string containing the name of the oauth flow

        """
        return self.default_oauth_flow

    def get_client_id( self ):
        """
            Get the current client id set in the credential.
    
            Returns: A string containing the client id
        """
        return self._client_id

    def get_username( self ):
        """
            Get the username of the user object

            Returns: A string containing the username of the credential
        """
        if self._username is not None:
            return self._username
        return None

    def has_ests( self ):
        """
            Check if the credential object has an ESTS cookie that has been set.

            Defaults to return False. Overridden in User class.

            Returns: bool
        """
        return False

    def get_id( self ):
        """
            Get the azol id of the object.
    
            Returns: string containing the azol id
        """
        return self._id

    def has_refresh_token(self):
        """
            Check if a user credential object has a refresh token saved.
        
            This will only return true for user objects that have a saved refresh token

            Returns: bool for whether a refresh token exists on the credential
        """
        return False
    


"""A module containing a key vault secret provider"""
import uuid
from azol.clients.key_vault_client import KeyVaultClient
from azol.credentials.user import User

class KeyVaultProvider:
    """
        A Key Vault secret provider for the azol token service
    """

    def __init__( self, username, key_vault_name, tenant_id, client_id=None, azol_id=None,
                  credential=None, cache_refresh_token=True ):
        self.key_vault_name = key_vault_name
        self.client_id=client_id
        self.tenant_id=tenant_id
        self.cache_refresh_token=cache_refresh_token
        if azol_id is None:
            azol_id = uuid.uuid4()
        self._id = azol_id
        if credential is None:
            self.credential = User(username=username)
        else:
            self.credential = credential
        self.kv_client = KeyVaultClient( key_vault_name, cred=self.credential,
                                       tenant=tenant_id, oauth_flow="device_code",
                                       use_persistent_cache=cache_refresh_token )

    def get_id( self ):
        """
            Get the azol id of the provider

            Returns:
                string containing the azol id
        """
        return self._id

    def get_secret( self, secret_reference ):
        """
            Get a secret using the provider

            Args:
                - secret_reference - A string containing the reference text for the secret
            
            Returns:
                string containing the secret
        """
        response = self.kv_client.get_secret( secret_reference )
        if response is None:
            return None
        return response
    


"""A module containing a data structure used in the initialization of OAuth login flows"""
from dataclasses import dataclass

@dataclass
class SPLoginModel( object ):
    """
        An object containing necessary data for SP-initiated login flows 
    """
    login_uri: str
    args: object